#ifndef __TS_H
#define __TS_H

/* A public domain implementation of some of ISO/IEC 13818-1 (MPEG2 system) */

#include <stdio.h> /* FILE* */
#include <string.h> /* memcpy */

/* to disable assert, define NDEBUG here on the command line */
/*#define NDEBUG*/
#include <assert.h>

/*
 * Contracts (pre/post conditions and invariants) are in 
 * function bodies in the .c file. Preconditions must 
 * be valid or checked explicitely before calling functions!
 * A concise intro to design by contract is at 
 * http://www.muc.de/~hoelzl/tools/dbc/dbc-intro.html
 * or feed "design by contract" to a search engine.
 */

/* Bit twiddling */

int int_lowest (int in, int bits);
int in_range (int in, int min, int max);
int in_pos_range (int in, int max);
int is_boolean (int in);

/* Buffer */

typedef struct buffer_st
{
	unsigned char *memory;
	int len;
} buffer;

int buffer_invariant (buffer *b);
void buffer_reset (buffer *b, unsigned char *ptr, int len);
void buffer_dump (buffer *b, FILE *f);

typedef struct buffer_iter_st 
{
	unsigned char *memory;
	int len;
	int byte_pos;
	int bit_pos;
} buffer_iter;

int buffer_iter_invariant (buffer_iter *b);
int buffer_iter_int_forth (buffer_iter *b, int bitlen);
int buffer_iter_bcd_forth (buffer_iter *b, int bitlen);
int buffer_iter_available_bits (buffer_iter *b);
int buffer_iter_available_bytes (buffer_iter *b);
void buffer_iter_zero (buffer_iter *b);
void buffer_iter_reset (buffer_iter *b, buffer *in);
void buffer_iter_skip_bytes (buffer_iter *b, int nbytes);
void buffer_iter_skip_bits (buffer_iter *b, int nbytes);
void buffer_memory_copy (buffer *dest, buffer *src);
void buffer_iter_memory_copy (buffer_iter *dest, buffer_iter *src, int nbytes);
void buffer_iter_truncate_bytes (buffer_iter *b, int nbytes);
char *buffer_iter_skip_bytes_or_error (buffer_iter *b, int nbytes, char *error);
char *buffer_iter_available_bytes_or_error (buffer_iter *b, int nbytes, char *error);
unsigned char *buffer_iter_memory_at (buffer_iter *b);
buffer buffer_iter_buffer_at (buffer_iter *b);
buffer buffer_iter_buffer_before (buffer_iter *b);
buffer_iter buffer_iter_buffer_iter_before (buffer_iter *b);
void buffer_iter_dump (buffer_iter *b, FILE *f);
void buffer_iter_xml_text_dump (buffer_iter *iter_in, FILE *f);

/* portable substitute for long long */

#define MPSYS_LONG_BASE 24

typedef struct mpsys_long_st
{
	int high;
	int low;
} mpsys_long;

mpsys_long mpsys_long_zero ();
void mpsys_long_print (mpsys_long *l, FILE *f);

/* clock reference */

typedef struct program_clock_reference_st
{
	mpsys_long base;
	int extension;
} program_clock_reference;

void program_clock_reference_print (program_clock_reference *c, FILE *f);

/* TS */

#define TS_LENGTH 188

typedef struct ts_adaptation_field_extension_st
{
	int ltw_flag;
	int ltw_valid_flag;
	int ltw_offset;
	
	int piecewise_rate_flag;
	int piecewise_rate;
	
	int seamless_splice_flag;
	int splice_type;
	mpsys_long dts_next_au;
	 
} ts_adaptation_field_extension;

typedef struct ts_adaptation_field_st
{
	int discontinuity_indicator;
	int random_access_indicator;
	int elementary_stream_priority_indicator;
	
	int pcr_flag;
	program_clock_reference program_clock_reference;
	
	int opcr_flag;
	program_clock_reference original_program_clock_reference;
	
	int splicing_point_flag;
	int splice_countdown;
	
	int transport_private_data_flag;
	int transport_private_data_length;
	/* ignored private data */
	
	int adaptation_field_extension_flag;
	ts_adaptation_field_extension adaptation_field_extension;
	
} ts_adaptation_field;

typedef struct ts_st
{
	int synch_byte;
	int transport_error_indicator;
	int payload_unit_start_indicator;
	int transport_priority;
	int pid;
	int transport_scrambling_control;
	int adaptation_field_control;
	int continuity_counter;
	
	ts_adaptation_field adaptation_field;
	
	buffer_iter data;
} ts;

int ts_invariant (ts *t);
int ts_is_ignorable (ts *t);
int ts_has_payload (ts *t);
int ts_starts_payload (ts *t);
int ts_in_sequence_counter (int first, int second);
int ts_in_sequence (ts *first, ts *second);
int ts_is_duplicate (ts *first, ts *second);
int ts_parsable (buffer *ts_in);
char *ts_parsable_error (buffer *ts_in);
void ts_parse (ts *t, buffer *ts_in);
void ts_print (ts *t, FILE *f);

/* TS payload */

typedef struct ts_payload_st {
	int id;
	int last_continuity;
	buffer payload_buffer;
	buffer_iter payload_writer;
	struct ts_payload_st *next;
} ts_payload;

int ts_payload_invariant (ts_payload *p);
void ts_payload_reset_payloads (ts_payload *head);
void ts_payload_reset_payload_writer (ts_payload *p);
void ts_payload_reset (ts_payload *p, int id, buffer *buf, ts_payload *next);
ts_payload *ts_payload_find (ts_payload *head, int id);
int ts_payload_has_last (ts_payload *p);
int ts_payload_ts_applies (ts_payload *p, ts *t);
char *ts_payload_ts_applies_error (ts_payload *p, ts *t);
void ts_payload_ts_apply (ts_payload *p, ts *t);
int ts_payload_process (ts_payload *ps, ts *t, buffer *out_payload);

/* Program info */

#define TS_PID_PAT 0
#define TS_PID_CAT 1
#define TS_PID_NIT 0x10
#define TS_PID_SDT 0x11
#define TS_PID_EIT 0x12
#define TS_PID_RST 0x13
#define TS_PID_TIME 0x14

#define PROGRAM_DESCRIPTOR_VIDEO 2
#define PROGRAM_DESCRIPTOR_AUDIO 3
#define PROGRAM_DESCRIPTOR_HIERARCHY 4
#define PROGRAM_DESCRIPTOR_REGISTRATION 5
#define PROGRAM_DESCRIPTOR_DATA_ALIGNMENT 6
#define PROGRAM_DESCRIPTOR_TARGET_BACKGROUND 7
#define PROGRAM_DESCRIPTOR_VIDEO_WINDOW 8
#define PROGRAM_DESCRIPTOR_CA 9
#define PROGRAM_DESCRIPTOR_LANGUAGE 10
#define PROGRAM_DESCRIPTOR_SYSTEM_CLOCK 11
#define PROGRAM_DESCRIPTOR_MULTIPLEX_BUFFER_UTILIZATION 12
#define PROGRAM_DESCRIPTOR_COPYRIGHT 13
#define PROGRAM_DESCRIPTOR_MAXIMUM_BITRATE 14
#define PROGRAM_DESCRIPTOR_PRIVATE_DATA 15
#define PROGRAM_DESCRIPTOR_SMOOTHING_BUFFER 16
#define PROGRAM_DESCRIPTOR_STD 17
#define PROGRAM_DESCRIPTOR_IBP 18

void ts_pointer_skip (buffer_iter *iter);
void ts_descriptor_forth (buffer_iter *iter);

char *ts_stream_type_name (int type);
char *ts_program_descriptor_language_audio_type_name (int type);

/* PES */

typedef struct pes_extension_st
{
	int program_packet_sequence_counter_flag;
	int program_packet_sequence_counter;
	int mpeg1_mpeg2_identifier;
	int original_stuff_length;
	
	int p_std_buffer_flag;
	int p_std_buffer_scale;
	int p_std_buffer_size;
} pes_extension;

typedef struct pes_st
{
	int stream_id;
	int packet_length;
	
	int scrambling_control;
	int priority;
	int data_alignment_indicator;
	int copyright;
	int original_or_copy;
	
	int pts_dts_flags;	
	mpsys_long pts;
	mpsys_long dts;
	
	int escr_flag;
	program_clock_reference escr;
	
	int es_rate_flag;
	int es_rate;
	
	int dsm_trick_mode_flag;
	/* pes_trick_mode trick_mode;*/ 
	
	int additional_copy_info_flag;
	int additional_copy_info;
	
	int crc_flag;
	int previous_packet_crc;
	
	int extension_flag;
	pes_extension extension;

	buffer_iter data;
} pes;

#define PES_PROGRAM_STREAM_MAP 0xBC /* 1011 1100 */
#define PES_PRIVATE_STREAM_1   0xBD /* 1011 1101 */
#define PES_PADDING_STREAM     0xBE /* 1011 1110 */
#define PES_PRIVATE_STREAM_2   0xBF /* 1011 1111 */

#define PES_ECM                0xF0 /* 1111 0000 */
#define PES_EMM                0xF1 /* 1111 0001 */
#define PES_DSMCC_STREAM       0xF2 /* 1111 0010 */
#define PES_ISO13552_STREAM    0xf3 /* 1111 0011 */

#define PES_ITU_A_STREAM       0xF4 /* 1111 0100 */
#define PES_ITU_B_STREAM       0xF5 /* 1111 0101 */
#define PES_ITU_C_STREAM       0xF6 /* 1111 0110 */
#define PES_ITU_D_STREAM       0xF7 /* 1111 0111 */
#define PES_ITU_E_STREAM       0xF8 /* 1111 1000 */

#define PES_ANCILLARY_STREAM   0xF9 /* 1111 1001 */

#define PES_PROGRAM_STREAM_DIRECTORY 0xFF

int pes_invariant (pes *p);
int pes_parsable (buffer *b);
char *pes_parsable_error (buffer *b);
void pes_parse (pes *p, buffer *p_in);
int pes_is_stream_video (pes *p);
int pes_is_stream_audio (pes *p);
int pes_stream_number (pes *p);
int pes_is_stream_other (pes *p);
void pes_print (pes *p, FILE *f);

/* Elementary stream: audio */

typedef struct es_audio_st {
	int version; /* 1, 2 */
	int layer; /* 1, 2, 3 */
	int has_crc;
	int bit_rate;
	int sample_rate;
	int has_padding;
	int channel_mode;
	int mode_extension;
	int copyright;
	int original;
	int emphasis;
	int crc;
	
	buffer_iter data;
} es_audio;

int es_audio_invariant (es_audio *a);
int es_audio_parsable (buffer *buf_in);
void es_audio_parse (es_audio *a, buffer *buf_in);
int es_audio_frame_length (es_audio *a);
int es_audio_length (es_audio *a);
void es_audio_print (es_audio *a, FILE *f);
char *es_audio_channel_mode_name (int channel_mode);
char *es_audio_emphasis_name (int mode);

#endif /* header */
