/* Audio elementary stream */

#include "mpsys.h"

static int es_audio_bitrates [16][2][3] = { /* bitrate * version * layer */
/*0000*/{{ -1,  -1,  -1}, { -1,  -1,  -1}}, /* ariable */

/*0001*/{{ 32,  32,  32}, { 32,   8,   8}},
/*0010*/{{ 64,  48,  40}, { 48,  16,  16}},
/*0011*/{{ 96,  56,  48}, { 56,  24,  24}},

/*0100*/{{128,  64,  56}, { 64,  32,  32}},
/*0101*/{{160,  80,  64}, { 80,  40,  40}},
/*0110*/{{192,  96,  80}, { 96,  48,  48}},

/*0111*/{{224, 112,  96}, {112,  56,  56}},
/*1000*/{{256, 128, 112}, {128,  64,  64}},
/*1001*/{{288, 160, 128}, {144,  80,  80}},

/*1010*/{{320, 192, 160}, {160,  96,  96}},
/*1011*/{{352, 224, 192}, {176, 112, 112}},
/*1100*/{{384, 256, 224}, {192, 128, 128}},

/*1101*/{{416, 320, 256}, {224, 144, 144}},
/*1110*/{{448, 384, 320}, {256, 160, 160}},
/*1111*/{{-2,   -2,  -2}, { -2,  -2,  -2}} /* illegal */
};

static int es_audio_samplerates [4][2] = {
/*00*/{44100, 22050}, /*11025*/
/*01*/{48000, 24000}, /*12000*/
/*10*/{32000, 16000}, /*8000*/
/*11*/{-1,-1}
};

/* invariant */
int es_audio_invariant (es_audio *a)
{
	return (a)
		&& in_range (a->version, 1, 2)
		&& in_range (a->layer, 1, 3)
		&& is_boolean (a->has_crc)
		&& is_boolean (a->has_padding)
		&& in_pos_range (a->channel_mode, 3)
		&& in_pos_range (a->mode_extension, 3)
		&& is_boolean (a->copyright)
		&& is_boolean (a->original)
		&& in_pos_range (a->emphasis, 3)
		;
}

/* whole packet length */
int es_audio_header_length (es_audio *a)
{
	assert (es_audio_invariant (a));

	return (4 + (a->has_crc * 2));
}

/* compute frame length */
int es_audio_frame_length (es_audio *a)
{
	assert (es_audio_invariant (a));
	
	if (a->layer == 1)
	{
		return (12 * (a->bit_rate * 1000) / a->sample_rate  + a->has_padding) * 4;
	}
	else
	{
		assert (a->layer >= 2);
		return (144 * a->bit_rate * 1000 / a->sample_rate + a->has_padding);
	}
}

/* stereo name */
char * es_audio_channel_mode_name (int channel_mode)
{
	assert (in_pos_range (channel_mode, 3));
	switch (channel_mode)
	{
		case 0: return "stereo";
		case 1: return "joint";
		case 2: return "dual";
		case 3: return "mono";
	}
	return "?";
}

/* emphasis name */
char *es_audio_emphasis_name (int mode)
{
	assert (in_pos_range (mode, 3));
	switch (mode)
	{
		case 0: return "none";
		case 1: return "50/15ms";
		case 2: return "";
		case 3: return "j.17";
	}
	return "?";
}

/* is well formed? */
int es_audio_parsable (buffer *buf_in)
{
	buffer_iter b;
	buffer_iter_reset (&b, buf_in);
	
	if (buffer_iter_available_bytes (&b) >= 4)
	{
		if (buffer_iter_int_forth (&b, 12) == 0xFFF) /* 1111 1111 1111 */
		{
			int has_crc;
			
			buffer_iter_skip_bits (&b, 1); /* version */
			if (buffer_iter_int_forth (&b, 2)) 
			{ /* layer not 0 */
				has_crc = !buffer_iter_int_forth (&b, 1);
				buffer_iter_skip_bits (&b, 16);
				if (has_crc)
				{
					if (buffer_iter_available_bytes (&b) >= 2)
					{
						return 1;
					}
				}
				else
				{
					return 1;
				}
			}
		}
	}
	return 0;
}

/* parse */
void es_audio_parse (es_audio *a, buffer *buf_in)
{
	buffer_iter b;
	buffer_iter_reset (&b, buf_in);
	
	assert  (es_audio_parsable (buf_in));
	buffer_iter_skip_bits (&b, 12);
	
	a->version = buffer_iter_int_forth (&b, 1) ? 1 : 2;
	 
	switch (buffer_iter_int_forth (&b, 2))
	{
		case 1: a->layer = 3; break;
		case 2: a->layer = 2; break;
		case 3: a->layer = 1; break;
	}
	
	a->has_crc = buffer_iter_int_forth (&b, 1) ? 0 : 1; /* not! */

	assert (!b.bit_pos);
	
	a->bit_rate = es_audio_bitrates [buffer_iter_int_forth (&b, 4)][a->version - 1][a->layer - 1];
	a->sample_rate = es_audio_samplerates [buffer_iter_int_forth (&b, 2)][a->version -1];
	
	a->has_padding = buffer_iter_int_forth (&b, 1);
	
	buffer_iter_skip_bits (&b, 1); /* private */
	
	assert (!b.bit_pos);

	a->channel_mode = buffer_iter_int_forth (&b, 2);
	a->mode_extension = buffer_iter_int_forth (&b, 2);

	a->copyright = buffer_iter_int_forth (&b, 1);
	a->original = buffer_iter_int_forth (&b, 1);
	
	a->emphasis = buffer_iter_int_forth (&b, 2);
	
	assert (!b.bit_pos);
	
	if (a->has_crc)
	{
		a->crc = buffer_iter_int_forth (&b, 16);
	}
	
	a->data = b; /*copy*/
	
	assert (es_audio_invariant (a));
}

/* print */
void es_audio_print (es_audio *a, FILE *f)
{
	assert (es_audio_invariant (a));
	
	fprintf (f, "ES: Audio MPEG%d(layer%d) %d Kbps %d Hz %s %d bytes",
		a->version,
		a->layer,
		a->bit_rate,
		a->sample_rate,
		es_audio_channel_mode_name (a->channel_mode),
		es_audio_frame_length (a));
		
	if (a->emphasis)
	{
		fprintf (f, " emph: %s", es_audio_emphasis_name (a->emphasis));
	}
	if (a->has_padding)
	{
		fprintf (f, " pad");
	}
	if (a->has_crc)
	{
		fprintf (f, " crc %x", a->crc);
	}
	if (a->copyright)
	{
		fprintf (f, " (c)");
	}
	if (a->original)
	{
		fprintf (f, " original");
	}
	fprintf (f, "\n");
}
