/* Packetized elementary stream (PES) */

#include "mpsys.h"

int pes_extension_invariant (pes_extension *pe)
{
	return (pe)
		&& is_boolean (pe->program_packet_sequence_counter_flag)
		&& (!(pe->program_packet_sequence_counter_flag) || 
			(in_pos_range (pe->program_packet_sequence_counter, 127)
			&& is_boolean (pe->mpeg1_mpeg2_identifier)
			&& in_pos_range (pe->original_stuff_length, 63)
			))
		&& is_boolean (pe->p_std_buffer_flag)
		&& (!(pe->p_std_buffer_flag) ||
			(is_boolean (pe->p_std_buffer_scale)
			&& in_pos_range (pe->p_std_buffer_size,(1<<13)-1)
			))
		;
}

int pes_invariant (pes *p)
{
	return (p)
		&& in_pos_range (p->stream_id, 255)
		&& (pes_is_stream_other (p) ||
			(in_pos_range (p->scrambling_control,3)
			&& is_boolean (p->priority)
			&& is_boolean (p->data_alignment_indicator)
			&& is_boolean (p->copyright)
			&& is_boolean (p->original_or_copy)
			
			&& in_pos_range (p->pts_dts_flags,3)
			&& is_boolean (p->escr_flag)
			&& is_boolean (p->es_rate_flag)
			&& is_boolean (p->dsm_trick_mode_flag)
			&& is_boolean (p->additional_copy_info_flag)
			&& is_boolean (p->crc_flag)
			&& is_boolean (p->extension_flag)
			
			&& (!(p->es_rate_flag) || is_boolean (p->es_rate))
			&& (!(p->additional_copy_info_flag) || in_pos_range (p->additional_copy_info,127))
			&& (!(p->crc_flag) || in_pos_range (p->previous_packet_crc,65535))
			&& (!(p->extension_flag) || pes_extension_invariant (&(p->extension)))
			))
		;
}

#define PES_AUDIO_STREAM_PREFIX 0x6 /* 110 */
#define PES_AUDIO_STREAM_BITS 5

#define PES_VIDEO_STREAM_PREFIX 0xE /* 1110 */
#define PES_VIDEO_STREAM_BITS 4


/* is this a stream type with another format? */
int pes_is_stream_id_other (int id)
{	
	return (id == PES_PROGRAM_STREAM_MAP)
		|| (id == PES_PADDING_STREAM)
		|| (id == PES_PRIVATE_STREAM_2)
		|| (id == PES_ECM)
		|| (id == PES_EMM)
		|| (id == PES_PROGRAM_STREAM_DIRECTORY)
		|| (id == PES_DSMCC_STREAM)
		|| (id == PES_ITU_E_STREAM)
		; 
}

static char *pes_stream_id_name (int id)
{
	if ((id >> PES_AUDIO_STREAM_BITS) == PES_AUDIO_STREAM_PREFIX)
	{
		return "audio";
	}
	if ((id >> PES_VIDEO_STREAM_BITS) == PES_VIDEO_STREAM_PREFIX)
	{
		return "video";
	}
		
	switch (id)
	{
		case PES_PROGRAM_STREAM_MAP:
			return "program-map";
		case PES_PADDING_STREAM:
			return "padding";
		case PES_PRIVATE_STREAM_1:
			return "private-1";
		case PES_PRIVATE_STREAM_2:
			return "private-2";
		case PES_ECM:
			return "ecm";
		case PES_EMM:
			return "emm";
		case PES_DSMCC_STREAM:
			return "dsmcc";
		case PES_ISO13552_STREAM:
			return "iso13552";
		case PES_ITU_A_STREAM:
			return "itu-A";
		case PES_ITU_B_STREAM:
			return "itu-B";
		case PES_ITU_C_STREAM:
			return "itu-c";
		case PES_ITU_D_STREAM:
			return "itu-d";
		case PES_ITU_E_STREAM:
			return "itu-e";
		case PES_ANCILLARY_STREAM:
			return "ancillary";
		case PES_PROGRAM_STREAM_DIRECTORY:
			return "program-directory";
			
		default:
			return "unknown-stream-type";
	}
	/* assert (result); */
}

/* is this a stream using another format? */
int pes_is_stream_other (pes *p)
{
	assert (p);
	return pes_is_stream_id_other (p->stream_id);
}

int pes_is_stream_audio (pes *p)
{
	assert (pes_invariant (p));
	return (p->stream_id >> PES_AUDIO_STREAM_BITS) == PES_AUDIO_STREAM_PREFIX;
}

int pes_is_stream_video (pes *p)
{
	assert (pes_invariant (p));
	return (p->stream_id >> PES_VIDEO_STREAM_BITS) ==PES_VIDEO_STREAM_PREFIX;
}

int pes_stream_number (pes *p)
{
	assert (pes_invariant (p));
	assert (pes_is_stream_audio (p) || pes_is_stream_video (p));
	if (pes_is_stream_audio (p))
	{
		return int_lowest (p->stream_id, PES_AUDIO_STREAM_BITS);
	}
	else
	{
		assert (pes_is_stream_video (p));
		return int_lowest (p->stream_id, PES_VIDEO_STREAM_BITS);
	}
}

static char *pes_parsable_impl (buffer *pes_in)
{
	char *result = 0;
	buffer_iter b;
	
	assert (buffer_invariant (pes_in));
	buffer_iter_reset (&b, pes_in);
	
	result = buffer_iter_available_bytes_or_error (&b, 6, "too short to be a PES packet");
	if (!result)
	{
		if (buffer_iter_int_forth (&b, 24)!=1) /* 0x000001 */
		{
			result = "prefix is not PES";
		}
		else
		{
			int stream_id = buffer_iter_int_forth (&b, 8);
			int packet_len = buffer_iter_int_forth (&b, 16);
			
			if (!pes_is_stream_id_other (stream_id))
			{
				if (packet_len > 0)
				{
					result = buffer_iter_available_bytes_or_error (&b, packet_len, "PES packet too short for content");
					if (!result)
					{
						buffer_iter_truncate_bytes (&b, packet_len);
					}
				} 
				/* else assert (packet_len == 0); -- implies len defined by transport stream */
				
				if (!result)
				{
					result = buffer_iter_available_bytes_or_error (&b, 3, "PES header too small for start");
				}
				
				if (!result)
				{
					if (buffer_iter_int_forth (&b,2) != 2) /* '10' */
					{
						result = "PES header marker missing";
					}
					else
					{
						int pts_dts = 0;
						int skip_bytes = 0;
						int extension = 0;
						
						buffer_iter_skip_bits (&b, 6);

						assert (!b.bit_pos);
						
						pts_dts = buffer_iter_int_forth (&b, 2);
						if (pts_dts == 2)
						{
							skip_bytes += 5;
						}
						else if (pts_dts == 3)
						{
							skip_bytes += 10; 
						}
						else if (pts_dts == 1)
						{
							result = "invalid pts/dts flag in PES header";
						}

						if (buffer_iter_int_forth (&b, 1))
						{ /* escr */
							skip_bytes += 6;
						}
						if (buffer_iter_int_forth (&b, 1))
						{ /* es */
							skip_bytes += 3;

						}
						if (buffer_iter_int_forth (&b, 1))
						{ /* DSM */
							skip_bytes += 1;
						}
						if (buffer_iter_int_forth (&b, 1))
						{ /* additional copyright */
							skip_bytes += 1;
						}
						if (buffer_iter_int_forth (&b, 1))
						{ /* crc */
							skip_bytes += 2;
						}
						extension = buffer_iter_int_forth (&b, 1);
						
						assert (!b.bit_pos);
						
						if (!result)
						{
							int header_data_len = buffer_iter_int_forth (&b, 8);
							result = buffer_iter_available_bytes_or_error (&b, header_data_len, "PES packet too short for header data");
							if (!result)
							{
								buffer_iter_truncate_bytes (&b, header_data_len);
							}
						}
						
						if (!result)
						{ /* skip all flags except extension (variable size) */
							result = buffer_iter_skip_bytes_or_error (&b, skip_bytes, "PES header too short for flags set");
						}
						
						if (!result && extension)
						{
							result = buffer_iter_available_bytes_or_error (&b, 1, "PES packet too short for extension header");
							if (!result)
							{
								int private = buffer_iter_int_forth (&b, 1);
								int pack_header_flag = buffer_iter_int_forth (&b, 1);
								int program_seq_counter_flag = buffer_iter_int_forth (&b, 1);
								int p_std_buffer_flag = buffer_iter_int_forth (&b, 1);
								int extension_flag_2 = 0;
								buffer_iter_skip_bits (&b, 3);
								extension_flag_2 = buffer_iter_int_forth (&b, 1);

								assert (!b.bit_pos);
								
								if (private)
								{
									result = buffer_iter_skip_bytes_or_error (&b, 16, 
											"PES packet too short for extension private field");
								}

								if (!result && pack_header_flag)
								{
									result = buffer_iter_available_bytes_or_error (&b, 1, "PES packet too short for extension pack length");
									if (!result)
									{
										result = buffer_iter_skip_bytes_or_error (&b, 
												buffer_iter_int_forth (&b, 8), /* pack len */ 
												"PES packet too short for extension pack");
									}			
								}

								if (!result && program_seq_counter_flag)
								{
									result = buffer_iter_skip_bytes_or_error (&b, 2, 
												"PES packet too short for extension seq counter");
								}				
								if (!result && p_std_buffer_flag)
								{
									result = buffer_iter_skip_bytes_or_error (&b, 2, 
												"PES packet too short for extension pack");
								}
								if (!result && extension_flag_2)
								{
									result = buffer_iter_available_bytes_or_error (&b, 1,
											"PES packet too short for extension flag 2");
									if (!result)
									{
										buffer_iter_skip_bits (&b, 1);
										result = buffer_iter_skip_bytes_or_error (&b,
											buffer_iter_int_forth (&b, 7), /* len */
											"PES packet too short for extension 2");
									}
								}
								assert (!b.bit_pos);
							}
						}
					}
				}
			}
		}
	}
	return result;
}

/* does this buffer contain a parsable PES header? */
int pes_parsable (buffer *b)
{
	assert (buffer_invariant (b));
	return pes_parsable_impl (b) == 0;
}

char *pes_parsable_error (buffer *b)
{
	assert (buffer_invariant (b));
	assert (!pes_parsable (b));
	return pes_parsable_impl (b);
	/* assert (result); */
}

/* get 32..30 marker 29..15 marker 14 ..0 */
static mpsys_long pes_marked_long_forth (buffer_iter *b)
{
	mpsys_long result;
	assert (MPSYS_LONG_BASE == 24);
	assert (buffer_iter_invariant (b));
	assert (buffer_iter_available_bits (b) >= 36); /* 3 + marker + 15 + marker + 15 + marker */
	
	result.high = buffer_iter_int_forth (b, 3) << 6;
	buffer_iter_skip_bits (b,1);
	result.high |= buffer_iter_int_forth (b, 6);
	
	result.low = buffer_iter_int_forth (b, 9) << 15;
	buffer_iter_skip_bits (b,1);
	result.low |= buffer_iter_int_forth (b, 15);
	
	buffer_iter_skip_bits (b,1);
	
	return result;
}

/* parse buffer containing valid PES packet */
void pes_parse (pes *p, buffer *p_in)
{
	buffer_iter b;

	assert (buffer_invariant (p_in));
	/* require */
	assert (pes_parsable (p_in)); /* this implies that all structural checks are done! */
	
	buffer_iter_reset (&b, p_in);
	
	buffer_iter_skip_bytes (&b, 3); /* prefix */
	p->stream_id = buffer_iter_int_forth (&b, 8);
	p->packet_length = buffer_iter_int_forth (&b, 16);
	if (p->packet_length > 0)
	{
		buffer_iter_truncate_bytes (&b, p->packet_length);
	}
	
	if (p->stream_id == PES_PADDING_STREAM)
	{
		buffer_iter_zero (&(p->data));
	}
	else if (pes_is_stream_other (p))
	{
		p->data = b; /* copy iter now at packet position, all packet is 'data' */
	}
	else
	{
		buffer_iter_skip_bits (&b, 2);
		p->scrambling_control = buffer_iter_int_forth (&b, 2);
		p->priority = buffer_iter_int_forth (&b, 1);
		p->data_alignment_indicator = buffer_iter_int_forth (&b, 1);
		p->copyright = buffer_iter_int_forth (&b, 1);
		p->original_or_copy = buffer_iter_int_forth (&b, 1);
		
		assert (!b.bit_pos);
		
		p->pts_dts_flags = buffer_iter_int_forth (&b, 2);
		p->escr_flag = buffer_iter_int_forth (&b, 1);
		p->es_rate_flag = buffer_iter_int_forth (&b, 1);
		p->dsm_trick_mode_flag = buffer_iter_int_forth (&b, 1);
		p->additional_copy_info_flag = buffer_iter_int_forth (&b, 1);
		p->crc_flag = buffer_iter_int_forth (&b, 1);
		p->extension_flag = buffer_iter_int_forth (&b, 1);
		
		assert (!b.bit_pos);

		{
			int header_data_length = buffer_iter_int_forth (&b, 8);
		
			/* init data, which is after header */
			p->data = b; /* copy iter */
			buffer_iter_skip_bytes (&(p->data), header_data_length); /* does not change b */
		}
		
		/* back to header */
		assert (!b.bit_pos);
		
		p->pts = mpsys_long_zero ();
		p->dts = mpsys_long_zero ();
		if (p->pts_dts_flags == 2)
		{
			buffer_iter_skip_bits (&b,4);
			p->pts = pes_marked_long_forth (&b);
		}
		if (p->pts_dts_flags == 3)
		{
			buffer_iter_skip_bits (&b,4);
			p->pts = pes_marked_long_forth (&b);
			buffer_iter_skip_bits (&b,4);
			p->dts = pes_marked_long_forth (&b);
		}
		
		assert (!b.bit_pos);
		
		if (p->escr_flag)
		{
			buffer_iter_skip_bits (&b,1);
			p->escr.base = pes_marked_long_forth (&b);
			p->escr.extension = buffer_iter_int_forth (&b, 9);
			buffer_iter_skip_bits (&b,1);
		}
		
		assert (!b.bit_pos);
		
		if (p->es_rate_flag)
		{
			buffer_iter_skip_bits (&b, 1);
			p->es_rate = buffer_iter_int_forth (&b, 22);
			buffer_iter_skip_bits (&b, 1);
		}
		
		assert (!b.bit_pos);
		
		if (p->dsm_trick_mode_flag)
		{
			buffer_iter_skip_bits (&b, 8);
		}
		
		assert (!b.bit_pos);
		
		if (p->additional_copy_info_flag)
		{
			buffer_iter_skip_bits (&b, 1);
			p->additional_copy_info = buffer_iter_int_forth (&b, 7);
		}
		
		assert (!b.bit_pos);
		
		if (p->crc_flag)
		{
			buffer_iter_int_forth (&b, 16);
		}

		assert (!b.bit_pos);
				
		if (p->extension_flag)
		{
			int private_data_flag = buffer_iter_int_forth (&b, 1);
			int pack_header_flag = buffer_iter_int_forth (&b, 1);
			p->extension.program_packet_sequence_counter_flag = buffer_iter_int_forth (&b, 1);
			p->extension.p_std_buffer_flag = buffer_iter_int_forth (&b, 1);
			buffer_iter_skip_bits (&b, 3); /* reserved */
			buffer_iter_skip_bits (&b, 1); /* extension_flag_2 */
			assert (!b.bit_pos);
						
			if (private_data_flag)
			{
				buffer_iter_skip_bytes (&b, 16);
			}
			assert (!b.bit_pos);
			
			if (pack_header_flag)
			{
				buffer_iter_skip_bytes (&b, buffer_iter_int_forth (&b, 8));
			}
			assert (!b.bit_pos);

			if (p->extension.program_packet_sequence_counter_flag)
			{
				buffer_iter_skip_bits (&b,1);
				p->extension.program_packet_sequence_counter_flag = buffer_iter_int_forth (&b,7);
				p->extension.mpeg1_mpeg2_identifier = buffer_iter_int_forth (&b,1);
				p->extension.original_stuff_length = buffer_iter_int_forth (&b,1);
			}
			assert (!b.bit_pos);

			if (p->extension.p_std_buffer_flag)
			{
				buffer_iter_skip_bits (&b, 2);
				p->extension.p_std_buffer_scale = buffer_iter_int_forth (&b, 1);
				p->extension.p_std_buffer_size = buffer_iter_int_forth (&b, 13);
			}
		}
	}
} 

void pes_print (pes *p, FILE *f)
{
	fprintf (f, "PES: ");
	
	/* stream id */
	if (pes_is_stream_audio (p))
	{
		fprintf (f, "audio %d ", pes_stream_number (p));
	}
	else if (pes_is_stream_video (p))
	{
		fprintf (f, "video %d ", pes_stream_number (p));
	}
	else
	{
		fprintf (f, "%s %x ", pes_stream_id_name (p->stream_id), p->stream_id);
	}
	
	fprintf (f, "data %d ", buffer_iter_available_bytes (&(p->data)));
	
	fprintf (f, "%s%s%s%s%s",
		p->scrambling_control ? "scrambled ":"",
		p->priority ? "priority ":"",
		p->data_alignment_indicator ? "data-align ":"",
		p->copyright ? "(c) ":"",
		p->original_or_copy ? "original ":"");
		
	if (p->pts_dts_flags)
	{
		fprintf (f, "pts ");
		mpsys_long_print (&(p->pts),f);
		fprintf (f, " dts ");
		mpsys_long_print (&(p->dts),f);
		fprintf (f, " ");
	}
	
	if (p->escr_flag)
	{
		fprintf (f, "es-clock ");
		program_clock_reference_print (&(p->escr), f);
		fprintf (f, " ");
	}
	
	if (p->es_rate_flag)
	{
		fprintf (f, "es-rate %d ", p->es_rate);
	}
	
	if (p->dsm_trick_mode_flag)
	{
		fprintf (f, "dsm-trick ");
	}
	
	if (p->additional_copy_info_flag)
	{
		fprintf (f, "copyinfo %d ", p->additional_copy_info);
	}
	
	if (p->crc_flag)
	{
		fprintf (f, "last-crc %04x", p->previous_packet_crc);
	}
	
	if (p->extension_flag)
	{
		pes_extension *pex = &(p->extension);
		
		if (pex->program_packet_sequence_counter_flag)
		{
			fprintf (f, "seq_counter %d %s %d",
				pex->program_packet_sequence_counter,
				pex->mpeg1_mpeg2_identifier ? "from-mpeg1": "from-ps",
				pex->original_stuff_length);
		}
		
		if (pex->p_std_buffer_flag)
		{
			fprintf (f, "std-buffer %s%d ", 
				pex->p_std_buffer_scale ? "scale ":"",
				pex->p_std_buffer_size); 
		}
	}
	
	fprintf (f, "\n");
}
