/* ----- Transport Stream (TS) */

#include "mpsys.h"
#include "mpsys_parser.h"

#define TS_SYNCH_BYTE 0x47


int ts_adaptation_field_extension_invariant (ts_adaptation_field_extension *afe)
{
	return (afe != 0); 	/* FIXME more */

}

int ts_adaptation_field_invariant (ts_adaptation_field *af)
{
	return (af)
		&& is_boolean (af->discontinuity_indicator)
		&& is_boolean (af->random_access_indicator)
		&& is_boolean (af->elementary_stream_priority_indicator)
		&& is_boolean (af->pcr_flag)
		&& is_boolean (af->opcr_flag)
		&& is_boolean (af->splicing_point_flag)
		&& is_boolean (af->transport_private_data_flag)
		&& is_boolean (af->adaptation_field_extension_flag)
		/* FIXME pcr opcr */
		&& ((af->splicing_point_flag == 0) || in_pos_range (af->splice_countdown, 255))
		&& ((af->adaptation_field_extension_flag == 0) 
			|| ts_adaptation_field_extension_invariant (&(af->adaptation_field_extension)));
}

static int ts_has_adaptation_field (ts *t)
{
	assert (t);
	/*assert (ts_invariant (t)); (commented to avoid loop) */
	return in_range (t->adaptation_field_control, 2, 3);
}

int ts_invariant (ts *t)
{
	return (t)
		&& (t->synch_byte == TS_SYNCH_BYTE)
		&& is_boolean (t->transport_error_indicator)
		&& is_boolean (t->payload_unit_start_indicator)
		&& is_boolean (t->transport_priority)
		&& in_pos_range (t->pid,(1<<13)-1)
		&& in_pos_range (t->transport_scrambling_control, 3)
		&& in_pos_range (t->adaptation_field_control, 3)
		&& in_pos_range (t->continuity_counter, 15)
		&& (!ts_has_adaptation_field (t) || ts_adaptation_field_invariant (&(t->adaptation_field)))
		;	 
}

/* should this packet be ignored */
int ts_is_ignorable (ts *t)
{
	assert (ts_invariant (t));
	return (t->pid == 0x1FFF) /* null */
		|| (t->adaptation_field_control == 0);
}

int ts_has_payload (ts *t)
{
	assert (ts_invariant (t));
	return  (t->adaptation_field_control == 1)
		|| (t->adaptation_field_control == 3);
}


int ts_starts_payload (ts *t)
{
	assert (ts_invariant (t));
	return (ts_has_payload (t) && t->payload_unit_start_indicator);
}

void ts_print (ts *t, FILE *f)
{
	assert (ts_invariant (t));
	
	fprintf(f, "TS: %d %s%s%s%s#%d %s",
			t->pid,
			t->transport_error_indicator ? "error " : "",
			t->payload_unit_start_indicator ? "start " : "",
			t->transport_priority ? "priority ":"",
			t->transport_scrambling_control ? "scrambled ":"",
			t->continuity_counter,
			!ts_has_payload(t) ? "no-payload ":"");
	if (ts_has_adaptation_field (t))
	{
		ts_adaptation_field *af = &(t->adaptation_field);

		fprintf(f, "[%s%s%s",
			af->discontinuity_indicator ? "discontin. ": "",
			af->random_access_indicator ? "rand ": "",
			af->elementary_stream_priority_indicator ? "es_priority ": "");

		if (af->pcr_flag)
		{
			fprintf (f, "pcr ");
			program_clock_reference_print (&(af->program_clock_reference), f);
		}
		if (af->opcr_flag)
		{
			fprintf (f, "orig-pcr ");
			program_clock_reference_print (&(af->original_program_clock_reference), f);
		}
		if (af->splicing_point_flag)
		{
			fprintf (f, "has-splice %d ", af->splice_countdown);
		}
		if (af->transport_private_data_flag)
		{
			fprintf (f, "has-priv ");
		}
		if (af->adaptation_field_extension_flag)
		{
			ts_adaptation_field_extension *afe = &(t->adaptation_field.adaptation_field_extension);
			if (afe->ltw_flag)
			{
				fprintf (f, "ltw %s%d ",
					afe->ltw_valid_flag ? "valid ":"",
					afe->ltw_offset);
			}
			if (afe->piecewise_rate_flag)
			{
				fprintf (f, "piecewise %x ",afe->piecewise_rate);
			}
			if (afe->seamless_splice_flag)
			{
				fprintf (f, "seamless-splice ");
				/* FIXME: fields...*/
			}
		}
		fprintf (f, "]");
	}
	fprintf (f, "\n");
}

int ts_in_sequence_counter (int first_continuity_counter, int second_continuity_counter)
{
	assert (first_continuity_counter >= 0 && first_continuity_counter <= 15);
	assert (second_continuity_counter >= 0 && second_continuity_counter <= 15);

	return ((first_continuity_counter + 1) == second_continuity_counter)
		|| ((first_continuity_counter == 15) && (second_continuity_counter == 0));
}

int ts_in_sequence (ts *first, ts *second)
{
	assert (ts_invariant (first));
	assert (ts_invariant (second));
	
	return ts_in_sequence_counter (first->continuity_counter, second->continuity_counter);
	/*FIXME adaptation.*/
}

int ts_is_duplicate (ts *first, ts *second)
{
	assert (ts_invariant (first));
	assert (ts_invariant (second));

	return (first->continuity_counter == second->continuity_counter)
		&& ts_has_payload (first) && ts_has_payload (second);
}

/* implementation for ts_parsable* that returns error message or void when OK */
static char *ts_parsable_impl (buffer *ts_in)
{
	char *result = 0; /* OK! */
	buffer_iter b;

	assert (ts_in);
	
	buffer_iter_reset (&b, ts_in);
	
	if (b.len < TS_LENGTH)
	{
		result = "buffer too small for TS packet";
	}
	else
	{
		if (buffer_iter_int_forth (&b, 8) != TS_SYNCH_BYTE)
		{
			result = "TS packet should start with synch byte";
		}
		else
		{
			int adapt_field_control;
			
			buffer_iter_skip_bits (&b, 18);
			adapt_field_control = buffer_iter_int_forth (&b, 2);
			buffer_iter_skip_bits (&b, 4);
			
			if (in_range (adapt_field_control, 2, 3))
			{ /* has adapt */
				int adapt_field_length = buffer_iter_int_forth (&b, 8);
				
				if ((adapt_field_control == 3) && !in_range (adapt_field_length, 0,182))
				{
					result = "invalid adapt field length with payload";
				}
				if ((adapt_field_control == 2) && (adapt_field_length != 183))
				{
					result = "invalid adapt field length when no payload";
				}
				
				if (adapt_field_length > 0)
				{ /* read flags or size checks */
					int pcr, opcr, splicing, tprivate, extension;
					int flagged_bytes = 0;
					
					buffer_iter_skip_bits (&b, 3);
					pcr = buffer_iter_int_forth (&b, 1);
					opcr = buffer_iter_int_forth (&b, 1);
					splicing = buffer_iter_int_forth (&b, 1);
					tprivate = buffer_iter_int_forth (&b, 1);
					extension = buffer_iter_int_forth (&b, 1);
					
					adapt_field_length--;
					
					if (pcr) { flagged_bytes += 6; }
					if (opcr) { flagged_bytes += 6; }
					if (splicing) { flagged_bytes += 1; }
					
					if (flagged_bytes > adapt_field_length)
					{
						result = "adaptation field too short for pcr";
					}
					else
					{
						adapt_field_length -= flagged_bytes;
						buffer_iter_skip_bytes (&b, flagged_bytes);
						
						if (tprivate)
						{
							int private_len = buffer_iter_int_forth (&b, 8);
							if (private_len > adapt_field_length)
							{
								result = "adaptation field too short for private transport data";
							}
							else
							{
								adapt_field_length -= private_len;
								buffer_iter_skip_bytes (&b, private_len);
								if (extension)
								{
									int extension_len = buffer_iter_int_forth (&b, 8);
									if (extension_len > adapt_field_length)
									{
										result = "adaptation field too short for extension";
									}
									else
									{
										flagged_bytes = 0;
										if (buffer_iter_int_forth (&b,1))
										{ /* ltw */ 
											flagged_bytes += 2;
										}
										if (buffer_iter_int_forth (&b,1))
										{ /* piecewise */ 
											flagged_bytes += 3;
										}
										if (buffer_iter_int_forth (&b,1))
										{
											flagged_bytes += 5;
										}
										buffer_iter_skip_bits (&b, 5);
										if (flagged_bytes > adapt_field_length)
										{
											result = "adaptation field too short for extension flags";
										} 

									}
								}
							}
						}
					}
				}
			}
		}
	}
	return result;
}
	
/* does the buffer contain parsable packet? */
int ts_parsable (buffer *ts_in)
{
	assert (ts_in);
	
	return (ts_parsable_impl (ts_in) == 0);
}

/* error message if unparsable */
char *ts_parsable_error (buffer *ts_in)
{
	assert (!ts_parsable (ts_in));
	
	return ts_parsable_impl (ts_in);
	/*assert (result);*/
}

void ts_parse (ts *t, buffer *ts_in)
{
	buffer_iter b;
	
	/* require */
	assert (ts_parsable (ts_in));
	
	buffer_iter_reset (&b, ts_in);
	assert (b.len >= TS_LENGTH);
	
	buffer_iter_zero (&(t->data)); /* in case no payload */

	t->synch_byte = buffer_iter_int_forth (&b, 8);
	t->transport_error_indicator = buffer_iter_int_forth (&b, 1);
	t->payload_unit_start_indicator = buffer_iter_int_forth (&b, 1);
	t->transport_priority = buffer_iter_int_forth (&b, 1);
	t->pid = buffer_iter_int_forth (&b, 13);
	t->transport_scrambling_control = buffer_iter_int_forth (&b, 2);
	t->adaptation_field_control = buffer_iter_int_forth (&b, 2);
	t->continuity_counter = buffer_iter_int_forth (&b, 4);

	assert (!b.bit_pos);
	
	/* adaptation field. */
	if ((t->adaptation_field_control==2) || (t->adaptation_field_control==3))
	{ 
		int adaptation_field_length = buffer_iter_int_forth (&b, 8);
		
		if (t->adaptation_field_control == 3)
		{ /* init payload */
			t->data = b; /* copy */
			buffer_iter_skip_bytes (&(t->data), adaptation_field_length);
		}
		
		if (adaptation_field_length > 0)
		{ /* 0 means length is 1 byte stuffing */ 
			ts_adaptation_field *af = &(t->adaptation_field);

			af->discontinuity_indicator = buffer_iter_int_forth (&b, 1);
			af->random_access_indicator = buffer_iter_int_forth (&b, 1);
			af->elementary_stream_priority_indicator = buffer_iter_int_forth (&b, 1);
			af->pcr_flag = buffer_iter_int_forth (&b, 1);
			af->opcr_flag = buffer_iter_int_forth (&b, 1);
			af->splicing_point_flag = buffer_iter_int_forth (&b, 1);
			af->transport_private_data_flag = buffer_iter_int_forth (&b, 1);
			af->adaptation_field_extension_flag = buffer_iter_int_forth (&b, 1);

			assert (!b.bit_pos);

			if (af->pcr_flag)
			{
				assert (MPSYS_LONG_BASE < 33);
				af->program_clock_reference.base.high = buffer_iter_int_forth (&b, 33-MPSYS_LONG_BASE);
				af->program_clock_reference.base.low = buffer_iter_int_forth (&b, MPSYS_LONG_BASE);
				buffer_iter_int_forth (&b, 6); /* reserved */
				af->program_clock_reference.extension = buffer_iter_int_forth (&b, 9);	
			}
			assert (!b.bit_pos);

			if (af->opcr_flag)
			{
				af->original_program_clock_reference.base.high = buffer_iter_int_forth (&b, 33-MPSYS_LONG_BASE);
				af->original_program_clock_reference.base.low = buffer_iter_int_forth (&b, MPSYS_LONG_BASE);
				buffer_iter_int_forth (&b, 6); /* reserved */
				af->original_program_clock_reference.extension = buffer_iter_int_forth (&b, 9);	
			}
			assert (!b.bit_pos);

			if (af->splicing_point_flag)
			{
				af->splice_countdown = buffer_iter_int_forth (&b, 8);
			}
			assert (!b.bit_pos);

			if (af->transport_private_data_flag)
			{
				int transport_private_data_length = buffer_iter_int_forth (&b, 8);
				buffer_iter_skip_bytes (&b, transport_private_data_length);
			}
			assert (!b.bit_pos);

			af->adaptation_field_extension.ltw_flag = 0;
			af->adaptation_field_extension.piecewise_rate_flag = 0;
			af->adaptation_field_extension.seamless_splice_flag = 0;
			
			if (af->adaptation_field_extension_flag)
			{
				ts_adaptation_field_extension *afext = &(af->adaptation_field_extension);
				int afext_length;
				
				afext_length = buffer_iter_int_forth (&b, 8);

				afext->ltw_flag = buffer_iter_int_forth (&b, 1);
				afext->piecewise_rate_flag = buffer_iter_int_forth (&b, 1);
				afext->seamless_splice_flag = buffer_iter_int_forth (&b, 1);
				buffer_iter_skip_bits (&b, 5); /* reserved */
				assert (!b.bit_pos);
				
				if (afext->ltw_flag)
				{
					afext->ltw_valid_flag = buffer_iter_int_forth (&b, 1);
					afext->ltw_offset = buffer_iter_int_forth (&b, 15);
				}
				assert (!b.bit_pos);

				if (afext->piecewise_rate_flag)
				{
					buffer_iter_skip_bits(&b, 2);
					afext->piecewise_rate = buffer_iter_int_forth (&b, 22);
				}
				if (afext->seamless_splice_flag)
				{
					afext->splice_type = buffer_iter_int_forth (&b, 4);
					afext->dts_next_au.high = (buffer_iter_int_forth (&b, 3) << 6);
					buffer_iter_skip_bits (&b, 1); /* marker */
					afext->dts_next_au.high |= buffer_iter_int_forth (&b, 6);
					afext->dts_next_au.low = (buffer_iter_int_forth (&b, 9) << 15);
					buffer_iter_skip_bits (&b, 1); /* marker */
					afext->dts_next_au.low |= buffer_iter_int_forth (&b, 15);
					buffer_iter_skip_bits (&b, 1); /* marker */
				}
			}
			assert (!b.bit_pos);
		}
	} /* has adapt field */
	else
	{
		/* b is at start of data */
		if (ts_has_payload (t))
		{
			t->data = b;
		}
	}
	assert (!b.bit_pos);


	assert (ts_invariant (t));
}

/* payload concatenator */

int ts_payload_invariant (ts_payload *p)
{
	return (p)
		&& (p->id >= 0)
		&& (in_range (p->last_continuity, -1, 15))
		&& (p->payload_buffer.memory == p->payload_writer.memory)
		;
}

/* reset payload writer to start of buffer */
void ts_payload_reset_payload_writer (ts_payload *p)
{
	assert (p);
	buffer_iter_reset (&(p->payload_writer), &(p->payload_buffer));
}

/* default values */
void ts_payload_reset (ts_payload *p, int id, buffer *buf, ts_payload *next)
{
	assert (p);
	
	p->id = id;
	p->last_continuity = -1;
	p->payload_buffer = *buf;
	ts_payload_reset_payload_writer (p);
	p->next = next;
	
	assert (ts_payload_invariant (p));
	assert (!ts_payload_has_last (p));
}

/* find a ts_payload in list for given PID */
ts_payload *ts_payload_find (ts_payload *head, int id)
{
	ts_payload *current = head; 
	while (current && (current->id != id))
	{
		current = current->next;
	}
	assert (!current || (current->id == id));
	return current;
}

/* Reset pending buffers, eg after an interruption or changing a filter */
void ts_payload_reset_payloads (ts_payload *head)
{
	ts_payload *current = head; 
	while (current)
	{
		ts_payload_reset_payload_writer (current);
		current = current->next;
	}
}


/* is there a valid last continuity (previous seen)? */
int ts_payload_has_last (ts_payload *p)
{
	assert (ts_payload_invariant (p));
	return (p->last_continuity>=0);
}

/* does this ts packet apply to this state (in sequence and payload) */
int ts_payload_ts_applies (ts_payload *p, ts *t)
{
	assert (ts_payload_invariant (p));
	assert (ts_invariant (t));
	
	return 
		(!ts_payload_has_last (p) /* no previous counter */
				|| ts_starts_payload (t) /* ignore counter when start */
				|| ts_in_sequence_counter (p->last_continuity, t->continuity_counter))
		&& (!ts_starts_payload (t) || (p->payload_writer.byte_pos == 0)) /* start implies writer at start */
		&& (buffer_iter_available_bytes (&(p->payload_writer)) >= buffer_iter_available_bytes (&(t->data)))
		;
}

/* error message for ts apply error */
char *ts_payload_ts_applies_error (ts_payload *p, ts *t)
{
	assert (p);
	assert (ts_invariant (t));
	if (ts_payload_has_last (p) && !ts_in_sequence_counter (p->last_continuity, t->continuity_counter))
	{
		return "not in sequence";
	}
	if (ts_starts_payload (t) && p->payload_writer.byte_pos > 0)
	{
		return "packet trailer before start";
	}
	if (buffer_iter_available_bytes (&(p->payload_writer)) < buffer_iter_available_bytes (&(t->data)))
	{
		return "payload buffer full!";
	}
	return "not in sequence";
}

/* apply incoming ts packet */
void ts_payload_ts_apply (ts_payload *p, ts *t)
{
	int data_len;
	
	assert (p);
	assert (ts_invariant (t));
	assert (ts_payload_ts_applies (p, t));
	
	if (ts_has_payload (t))
	{
		data_len = buffer_iter_available_bytes (&(t->data));
		buffer_iter_memory_copy (&(p->payload_writer), &(t->data), data_len);
		buffer_iter_skip_bytes (&(p->payload_writer), data_len);
	}
	p->last_continuity = t->continuity_counter;
	assert (ts_payload_has_last (p));
}

/* add a packet to the corresponding payload */
int ts_payload_process (ts_payload *states, ts *t, buffer *out_payload)
{
	ts_payload *state;
	int out_has_payload = 0;
	assert (ts_invariant (t));
	
	state = ts_payload_find (states, t->pid);
	if (state)
	{
		/* is this the first TS of a payload? */
		if (ts_starts_payload (t))
		{
			/* process previous payload if any */
			if(state->payload_writer.byte_pos > 0)
			{
				/* there was some previously processed payload */
				buffer from = buffer_iter_buffer_before (&state->payload_writer);
				buffer_memory_copy (out_payload, &from);
				
				out_has_payload = 1;
			}
			/* resets write */
			ts_payload_reset_payload_writer (state);
		}
		
		/* add to payload */
		if (ts_payload_ts_applies (state, t))
		{
			ts_payload_ts_apply (state, t);
		}
		/* else verbose message? */
	}
	return out_has_payload;
}

/* Program Information */

/* pointer: skip length byte before table */
void ts_pointer_skip (buffer_iter *iter)
{
	if (buffer_iter_available_bytes (iter) > 0)
	{
		int pointer = buffer_iter_int_forth (iter, 8);
		if (buffer_iter_available_bytes (iter) >= pointer)
		{
			buffer_iter_skip_bytes (iter, pointer);
		}
	}
}

/* for descriptor loops */
void ts_descriptor_forth (buffer_iter *iter)
{
	assert (buffer_iter_invariant (iter));
	assert (iter->bit_pos == 0);
	
	if (buffer_iter_available_bytes (iter) >= 2)
	{
		int tag = buffer_iter_int_forth (iter, 8);
		int len = buffer_iter_int_forth (iter, 8);

		if (len > buffer_iter_available_bytes (iter))
		{
			fprintf (stderr, "descriptor %d has invalid length %d\n", tag, len);
			buffer_iter_skip_bytes (iter, buffer_iter_available_bytes (iter));
		}
		else
		{
			buffer_iter_skip_bytes (iter, len);
		}
	}
	else
	{
		/* tag/length does not fit, go to end */
		buffer_iter_skip_bytes (iter, buffer_iter_available_bytes (iter));
	}
}

/* names */

char *ts_stream_type_name (int type)
{
	switch (type)
	{
		case 0: return "reserved";
		case 1: return "mpeg1-video";
		case 2: return "mpeg2-video";
		case 3: return "mpeg1-audio";
		case 4: return "mpeg2-audio";
		case 5: return "private";
		case 6: return "pes-private";
		case 7: return "MHEG";
		case 8: return "dsm";
		case 9: return "H222.1";
		case 0xA: return "13818-6-A";
		case 0xB: return "13818-6-B";
		case 0xC: return "13818-6-C";
		case 0xD: return "13818-6-D";
		case 0xE: return "13818-1-aux";
	}
	return "unknown";
}

char *ts_program_descriptor_language_audio_type_name (int type)
{
	switch (type)
	{
		case 0: return "undefined";
		case 1: return "clean-effects";
		case 2: return "hearing-impaired";
		case 3: return "visually-impaired";
	}
	return "reserved";
}
