#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "mpsys.h"
#include "mpsys_parser.h"
#include "dvbsi_parser.h"

#include "ts_config.h"
#include "ts_info.h"

#define PES_MAX_LENGTH 100000 /* completely arbitrary */

int is_output_info (config *cfg)
{
	assert (cfg);
	return (cfg->out == OUT_INFO_XML) ||(cfg->out == OUT_INFO_TEXT);
}

/* is this PES packet accepted with current config? */
int filter_pes_accept (pes *p, config *cfg)
{
	assert (pes_invariant (p));
	assert (cfg);
	
	if (cfg->filter_audio && cfg->filter_video)
	{
		return pes_is_stream_audio (p) || pes_is_stream_video (p);
	}
	else if (cfg->filter_audio)
	{
		return pes_is_stream_audio (p);
	}
	else if (cfg->filter_video)
	{
		return pes_is_stream_video (p);
	}
	return 1; /* accept everything by default */
}

/* is this TS packet accepted with current config? */
int filter_ts_accept (ts *t, config *cfg, struct info_state_st *info)
{
	assert (ts_invariant (t));
	assert (cfg);
	
	if (is_output_info (cfg))
	{
		return t->pid == info_pid (info);
	}
	if (cfg->filter_by_pid)
	{
		return t->pid == cfg->filtered_pid;
	}
	return 1; /* accept everything by default */
}

/* process ES audio */
void process_es_audio (pes *p)
{
	es_audio a;
	buffer_iter data_iter = p->data; /* copy */
	
	assert (pes_is_stream_audio (p));
	
	do{
		buffer b = buffer_iter_buffer_at (&data_iter);
		if (es_audio_parsable (&b))
		{
			es_audio_parse (&a, &b);
			es_audio_print (&a, stderr);
			buffer_iter_skip_bytes (&data_iter, es_audio_frame_length(&a));
		}
		else
		{
			fprintf (stderr, "ES: not an audio ES frame\n");
			buffer_iter_dump (&data_iter, stderr);
			break;
		}
	} while (buffer_iter_available_bytes (&data_iter)>0);
}

void buffer_out (buffer *b, config *cfg)
{
	assert (buffer_invariant (b));
	assert (cfg);
	
	if (cfg->out_hex)
	{
		buffer_dump (b, stderr);
	}
	else
	{
		fwrite (b->memory, b->len, 1, cfg->fd_out);
	}
}

/* test pes prefix */
int is_pes (buffer *pes_in)
{
	buffer_iter iter;
	buffer_iter_reset (&iter, pes_in);
	return buffer_iter_int_forth (&iter, 24) == 1;
}

/* process incoming PES buffer */
void process_pes (buffer *pes_in, config *cfg)
{
	if (pes_parsable (pes_in))
	{	
		pes p;
		pes_parse (&p, pes_in);
		
		if (filter_pes_accept (&p, cfg))
		{
			if (cfg->dump_pes)
			{
				pes_print (&p, stderr);
			}
			if (cfg->dump_es)
			{
				if (pes_is_stream_audio (&p))
				{
					process_es_audio (&p);
				} /* else FIXME: video */
			}
			if (cfg->out == OUT_PES)
			{
				buffer_out (pes_in, cfg);
			}
			if (cfg->out == OUT_ES)
			{
				buffer data = buffer_iter_buffer_at (&(p.data));
				buffer_out (&data, cfg);
			}
		}
	}
	else
	{
		if (is_pes (pes_in) || cfg->verbose)
		{
			fprintf (stderr, "Invalid PES packet %s\n", pes_parsable_error (pes_in));
		}
	}
}

static void *malloc_or_exit (int size)
{
	void *result = malloc (size);
	if (!result)
	{
		fprintf (stderr, "Out of memory!\n");
		exit (2);
	}
	assert (result);
	return result;
}
	
ts_payload *ts_payload_add (ts_payload *head, int id)
{
	ts_payload *result;
	buffer result_buffer;
	assert (!ts_payload_find (head, id));
	
	result = malloc_or_exit (sizeof(ts_payload));
	buffer_reset (&result_buffer, malloc_or_exit (PES_MAX_LENGTH), PES_MAX_LENGTH);
	ts_payload_reset (result, id, &result_buffer, head);
	return result;
}

/* descriptors */

void process_descriptors (buffer_iter *iter_in)
{
	buffer_iter iter;
	for (iter = *iter_in; buffer_iter_available_bytes (&iter)>2; ts_descriptor_forth (&iter))
	{
		buffer buf = buffer_iter_buffer_at (&iter); /* start of desc */
				
		fprintf (stderr,"--");
		if (ts_descriptor_parsable (&buf))
		{
			ts_descriptor_parse_and_print (&buf, stderr);
		}
		else if (dvb_descriptor_parsable (&buf))
		{
			dvb_descriptor_parse_and_print (&buf, stderr);
		}
		else
		{
			buffer_iter desc_iter = iter;
			int tag = buffer_iter_int_forth (&desc_iter, 8);
			int len = buffer_iter_int_forth (&desc_iter, 8);
			buffer_iter_truncate_bytes (&desc_iter, len);
			fprintf (stderr, "unknown descriptor %x:", tag);
			buffer_iter_dump (&desc_iter, stderr);
		}
	}
}

/* nicer than the generated _print */
void ts_program_association_pretty_print (ts_program_association *as, FILE *f)
{
	int i;
	
	assert (as);
	
	fprintf (f, "PAT: %d v%d%s %d/%d - ",
		as->transport_stream_id,
		as->version_number,
		as->current ? " current":"",
		as->section_number,
		as->last_section_number);
	for (i = 0; i < as->programs_count; i++)
	{
		int pn = as->programs[i].program_number;
		int pid =  as->programs[i].pid;
		if (pn == 0)
		{
			fprintf (f, "net_info %d ", pid);
		}
		else
		{
			fprintf (f, "%d:%d ", pn, pid);
		}
	}
	fprintf (f, "\n");
}

void process_program_map (buffer *buf)
{
	ts_program_map pm;
	int i;
	
	assert (buffer_invariant (buf));
	assert (ts_program_map_parsable (buf));
	
	ts_program_map_parse (&pm, buf);
	fprintf (stderr, "PGM-MAP: #%d v%d%s %d/%d pcr %d\n",
		pm.program_number,
		pm.version_number,
		!pm.current ? " next-not-current":"",
		pm.section_number, pm.last_section_number,
		pm.pcr_pid);

	process_descriptors (&(pm.descriptors));

	for(i = 0; i < pm.streams_count; i++)
	{
		fprintf (stderr, "-stream: type %s id %d\n", 
				ts_stream_type_name (pm.streams[i].stream_type),
				pm.streams[i].es_pid);
		process_descriptors (&(pm.streams[i].descriptors));
	}
}

void process_service_description (buffer *buf)
{
	service_description sd;
	int i;

	assert (buffer_invariant (buf));
	assert (service_description_parsable (buf));
	
	service_description_parse (&sd, buf);
	fprintf (stderr, "PGM-SERVICE:%s #%d v%d%s %d/%d orig-net-id %d\n",
		sd.table_id == 0x46 ? " other-ts":"",
		sd.transport_stream_id,
		sd.version_number,
		!sd.current_next_indicator ? " next-not-current":"",
		sd.section_number, sd.last_section_number,
		sd.original_network_id);
	for (i = 0; i < sd.services_count; i++)
	{
		fprintf (stderr, "-service: %d%s%s running %d%s\n",
			sd.services[i].service_id,
			sd.services[i].eit_schedule_flag ? " has-eit-schedule":"",
			sd.services[i].eit_present_following_flag ? " eit-pres-follow":"",
			sd.services[i].running_status,
			sd.services[i].free_ca_mode ? " scrambled":"");
		process_descriptors (&(sd.services[i].descriptors));
	}
}


void process_payload_info (int payload_id, buffer *payload, config *cfg)
{
	if (cfg->dump_info)
	{
		buffer_iter payload_iter;
		buffer buf;

		buffer_iter_reset (&payload_iter, payload);
		ts_pointer_skip (&payload_iter);

		buf = buffer_iter_buffer_at (&payload_iter);

		if (payload_id == TS_PID_PAT)
		{
			if(ts_program_association_parsable (&buf))
			{
				ts_program_association as;
				ts_program_association_parse (&as, &buf);
				ts_program_association_pretty_print (&as, stderr);
			}
			else
			{
				fprintf (stderr, "can't parse program association: %s\n", ts_program_association_parsable_error (&buf));
			}
		}
		else if (payload_id == TS_PID_SDT)
		{
			if (service_description_parsable (&buf))
			{
				process_service_description (&buf);
			}
			else if (bouquet_association_parsable (&buf))
			{
				bouquet_association_parse_and_print (&buf, stderr);
			}
			else
			{
				fprintf (stderr, "can't parse service description\n");
			}
		}
		else if (payload_id == TS_PID_NIT)
		{
			if (network_information_parsable (&buf))
			{
				network_information_parse_and_print (&buf, stderr);
			}
			else
			{
				fprintf (stderr, "can't parse network info %s\n", network_information_parsable_error (&buf));
			}
		}
		else if (payload_id == TS_PID_EIT)
		{
			if (event_information_parsable (&buf))
			{
				event_information_parse_and_print (&buf, stderr);
			}
			else
			{
				fprintf (stderr, "can't parse event info\n");
			}
		}
		else if(payload_id == TS_PID_RST)
		{
			if (running_status_parsable (&buf))
			{
				running_status_parse_and_print (&buf, stderr);
			}
			else
			{
				fprintf (stderr, "can't parse running status info\n");
			}
		}
		else if (payload_id == TS_PID_TIME)
		{
			if(time_date_section_parsable (&buf))
			{
				time_date_section_parse_and_print (&buf, stderr);
			}
			else if (time_offset_table_parsable (&buf))
			{
				time_offset_table_parse_and_print (&buf, stderr);
			}
			else
			{
				fprintf (stderr, "can't parse time date table\n");
			}
		}
		else
		{
			if (ts_program_map_parsable (&buf))
			{
				process_program_map (&buf);
			}
		}
	}
}

void reset_config_for_info (struct info_state_st *info, config *cfg)
{
	info->do_nit = cfg->out_info_with_nit;
	if (cfg->out == OUT_INFO_XML)
	{
		set_info_xml (info);
	}
}

/*#define DEV_NODE "/dev/twinhan3"*/
int main (int argc, char **argv)
{
	config cfg;
	
	unsigned char buf_memory[TS_LENGTH];
	buffer buf;

	struct info_state_st info;
	
	ts_payload *states = 0;
	
	if (config_parse (&cfg, argc, argv))
	    return -1;

	/* config for info output */
	reset_info (&info);
	reset_config_for_info (&info, &cfg);
	
	buffer_reset (&buf, buf_memory, TS_LENGTH);
	
	int nDevHandle = open(cfg.dev, O_RDONLY);
    if (nDevHandle < 0) {
        fprintf (stderr, "Can't find twinhan device!\n");
        return nDevHandle;
    }
    
    DEMULTIPLEXER_VALUE DemultiplexerValue;
    DemultiplexerValue.StreamType = TH_ALL_PID_TS_PACKET;	
    if (ioctl(nDevHandle, READ_ENABLE, &DemultiplexerValue)) {
        fprintf (stderr, "ioctl READ_ENABLE err !!\n");
        return -1;
    }    
    
    while (read(nDevHandle, buf.memory, TS_LENGTH))
	{
		if (ts_parsable (&buf))
		{
			ts packet;
			ts_parse (&packet, &buf);
			
			if (filter_ts_accept (&packet, &cfg, &info))
			{
				unsigned char payload_memory[PES_MAX_LENGTH];
				buffer payload;
				buffer_reset (&payload, payload_memory, sizeof (payload_memory));
				
				if (cfg.dump_ts)
				{
					ts_print (&packet, stderr);
				}

				if (cfg.out == OUT_TS)
				{
					buffer_out (&buf, &cfg);
				}
				
				/* create state if none for this PID */
				if(! ts_payload_find (states, packet.pid))
				{ 
					states = ts_payload_add (states, packet.pid);
				}
				
				if (ts_payload_process (states, &packet, &payload))
				{
					/* process previous payload if any */
					if (is_output_info (&cfg))
					{
						/* process_info may change the filter, for this 
						reason the currently partially filled buffers may be 
						from before an interruption via the filter and be 
						out of synch, so reset them */
						ts_payload_reset_payloads (states);
						
						process_info (&info, packet.pid, &payload);
						if (info.status == TS_INFO_FINISHED)
						{
						    close(nDevHandle); 
							return 0;
						}
					}
					else 
					{
						process_pes (&payload, &cfg);
						process_payload_info (packet.pid, &payload, &cfg);
					}
				}
			} /* filter ts */
		}
		else
		{
			fprintf (stderr, "Not a TS packet: %s\n", ts_parsable_error (&buf));
		}
	}
	fprintf (stderr, "No more packets.\n");
	close(nDevHandle); 
	return 0;
}
